"use strict";
/// <reference types="node" />
Object.defineProperty(exports, "__esModule", { value: true });
exports.MemorySlidingWindowRateLimiter = void 0;
const events_1 = require("events");
class MemorySlidingWindowRateLimiter extends events_1.EventEmitter {
    constructor(options = {}) {
        super();
        this.options = options;
        this.buckets = {};
        this.timers = {};
        this.interval = options.interval || 60000;
    }
    async cancel(key, token) {
        this.bucketExpireNow(key);
        const usage = this.buckets[key].length;
        let position;
        while ((position = this.buckets[key].indexOf(token)) !== -1) {
            this.buckets[key].splice(position, 1);
        }
        const usage2 = this.buckets[key].length;
        return {
            canceled: usage - usage2,
        };
    }
    async check(key, limit) {
        const now = this.bucketExpireNow(key);
        const usage = this.buckets[key].length;
        const result = { usage };
        if (usage && usage >= limit) {
            const reset = this.bucketResetValue(key, limit, now);
            if (reset)
                result.reset = reset;
        }
        return result;
    }
    async reserve(key, limit) {
        const now = this.bucketExpireNow(key);
        const usage = this.buckets[key].length;
        if (usage && usage >= limit) {
            const reset = this.bucketResetValue(key, limit, now);
            const result = { usage };
            if (reset)
                result.reset = reset;
            return result;
        }
        else {
            this.buckets[key].push(now);
            if (this.timers[key]) {
                clearTimeout(this.timers[key]);
            }
            this.timers[key] = setTimeout(() => {
                delete this.buckets[key];
                delete this.timers[key];
            }, this.interval).unref();
            const result = { usage: usage + 1 };
            result.token = now;
            if (usage + 1 >= limit) {
                const reset = this.bucketResetValue(key, limit, now);
                if (reset)
                    result.reset = reset;
            }
            return result;
        }
    }
    destroy() {
        for (const key of Object.keys(this.timers)) {
            clearTimeout(this.timers[key]);
        }
    }
    bucketResetValue(key, limit, now) {
        const usage = this.buckets[key].length;
        const oldest = this.buckets[key][usage - limit];
        return oldest ? oldest + this.interval - now : undefined;
    }
    bucketExpireNow(key) {
        const now = new Date().getTime();
        this.buckets[key] = this.buckets[key] ? this.buckets[key].filter(ts => now - ts < this.interval) : [];
        return now;
    }
}
exports.MemorySlidingWindowRateLimiter = MemorySlidingWindowRateLimiter;
exports.default = MemorySlidingWindowRateLimiter;
//# sourceMappingURL=memory-sliding-window-rate-limiter.js.map