"use strict";
/// <reference types="node" />
Object.defineProperty(exports, "__esModule", { value: true });
exports.RedisSlidingWindowRateLimiter = void 0;
const tslib_1 = require("tslib");
const events_1 = require("events");
const fs_1 = tslib_1.__importDefault(require("fs"));
const path_1 = tslib_1.__importDefault(require("path"));
const IORedis = require("ioredis");
const timeout_error_1 = require("./timeout-error");
const LUA = {
    cancel: fs_1.default.readFileSync(path_1.default.join(__dirname, "../src/redis/cancel.lua"), "utf8"),
    check: fs_1.default.readFileSync(path_1.default.join(__dirname, "../src/redis/check.lua"), "utf8"),
    reserve: fs_1.default.readFileSync(path_1.default.join(__dirname, "../src/redis/reserve.lua"), "utf8"),
};
class RedisSlidingWindowRateLimiter extends events_1.EventEmitter {
    constructor(options = {}) {
        super();
        this.options = options;
        this.interval = Number(options.interval) || 60000;
        this.operationTimeout = options.operationTimeout || 0;
        if (!options.redis || typeof options.redis === "string") {
            this.redis = new IORedis({
                host: options.redis,
                retryStrategy: _times => 1000,
                maxRetriesPerRequest: 1,
            });
        }
        else {
            this.redis = options.redis;
        }
        this.redis.defineCommand("limiter_cancel", { lua: LUA.cancel, numberOfKeys: 1 });
        this.redis.defineCommand("limiter_check", { lua: LUA.check, numberOfKeys: 1 });
        this.redis.defineCommand("limiter_reserve", { lua: LUA.reserve, numberOfKeys: 1 });
    }
    cancel(key, token) {
        return this.promiseWithTimeoutError(this.redis.limiter_cancel(key, token)).then(canceled => ({ canceled }));
    }
    check(key, limit) {
        return this.promiseWithTimeoutError(this.redis.limiter_check(key, this.interval, limit)).then(values => {
            const result = {
                usage: values[0],
            };
            if (values[1])
                result.reset = Math.ceil(values[1] / 1000);
            return result;
        });
    }
    reserve(key, limit) {
        return this.promiseWithTimeoutError(this.redis.limiter_reserve(key, this.interval, limit)).then(values => {
            const result = {
                usage: values[1],
            };
            if (values[0])
                result.token = values[0];
            if (values[2])
                result.reset = Math.ceil(values[2] / 1000);
            return result;
        });
    }
    destroy() {
        if (!this.options.redis || typeof this.options.redis === "string") {
            try {
                this.redis.quit().catch(() => {
                    this.redis.disconnect();
                });
            }
            catch (e) {
                this.redis.disconnect();
            }
        }
    }
    async promiseWithTimeoutError(operationPromise) {
        if (this.operationTimeout) {
            let timer;
            const timeoutPromise = new Promise((_resolve, reject) => {
                timer = setTimeout(() => {
                    reject(new timeout_error_1.TimeoutError());
                }, this.operationTimeout).unref();
            });
            const result = await Promise.race([operationPromise, timeoutPromise]);
            if (timer)
                clearTimeout(timer);
            return result;
        }
        else {
            return operationPromise;
        }
    }
}
exports.RedisSlidingWindowRateLimiter = RedisSlidingWindowRateLimiter;
exports.default = RedisSlidingWindowRateLimiter;
//# sourceMappingURL=redis-sliding-window-rate-limiter.js.map